#include <EEPROM.h>
#include <Wire.h>
#include <Adafruit_PWMServoDriver.h>
#include <avr/wdt.h>

// objekt pro ovladani PWM multiplexeru
Adafruit_PWMServoDriver driverPCA = Adafruit_PWMServoDriver(0x40);


#define GRID 1
#define FVE 2
#define NC 0
#define NEMENIT 0

// struktura pro ukladani stavu prepinacu
typedef struct {
  const unsigned char   assoc;            // asociace s rele termostatu 0-11
  unsigned char         poloha;           // aktualni poloha GRID nebo FVE, obnoveno z EEPROM
  unsigned char         poloha_poz;       // pozadovana poloha GRID, FVE nebo NEMENIT
  unsigned char         moving;           // zamek pro zapis pozadovane polohy od mastera
  const unsigned int    uhel_grid;        // uhel pro sepnuti k siti
  const unsigned int    uhel_fve;         // uhel pro sepnuti k fve
  const unsigned int    uhel_nc;          // uhel stredni polohy
  unsigned int          uhel_poz;         // pozadovany uhel natoceni serva
  unsigned int          uhel;             // aktualni uhel natoceni serva
  unsigned int          counter_nc;       // prodleva na stredni poloze
} TSERVO;

// pocet servo-prepinacu
#define POCETSRV  2

// seznam servoprepinacu
TSERVO srv[POCETSRV] = {
  {1, GRID, NEMENIT, false, 496, 180, 338, 0, 0, 200},
  {2, GRID, NEMENIT, false, 180, 496, 338, 0, 0, 200}
};

// navratovy typ detekce zmeny
typedef struct {
  unsigned char         foundRequest;
  unsigned char         servoID;
} TRES;

// zapnuti / vypnuti napajeni mppt regulatoru
#define FVEPOWER_IO   A1
#define FVEPOWER_LED  8
#define FVEPOWER_OFF  digitalWrite(FVEPOWER_IO, LOW); digitalWrite(FVEPOWER_LED, LOW);
#define FVEPOWER_ON   digitalWrite(FVEPOWER_IO, HIGH); digitalWrite(FVEPOWER_LED, HIGH)

// zapnuti / vypnuti napajeni servomotorku
#define SRVPOWER_IO   A0
#define SRVPOWER_LED  7
#define SRVPOWER_OFF  digitalWrite(SRVPOWER_IO, LOW); digitalWrite(SRVPOWER_LED, LOW);
#define SRVPOWER_ON   digitalWrite(SRVPOWER_IO, HIGH); digitalWrite(SRVPOWER_LED, HIGH);

// pin na vypnuti napajeni mppt, reset vyroby v noci
#define POWEROFFMPPT_IO    11

// pin na povoleni/zakazani prepnuti na FVE
#define DISFVE_IO     12

// bool na povoleni/zakazani prepnuti na FVE
unsigned char fveSwitchEnabled = 0;

// bool na vynucene vypnuti FVE pro ucely resetu vyroby
unsigned char fvePowerOff = 0;

// pro casovadlo
unsigned long lastTime10ms = 0;

// adresa zacatku ukladani poloh do eeprom
#define EEADDR  1

// stavovy prostor
#define STATE1   1
#define STATE2   2
#define STATE3   3
#define STATE4   4
#define STATE5   5
#define STATE6   6
#define STATE7   7
#define STATE8   8
#define INIT 100
#define STBY 101
unsigned char state = INIT;
unsigned int stateCounter = 0;




// -- setup funkce --
void setup() {

  // zapnuti psa
  wdt_enable(WDTO_8S);

  // nastaveni komunikace via RS485 bus
  commSetup();

  Serial.println("Start...");

  // pin na povoleni/zakazani prepnuti na FVE
  pinMode(DISFVE_IO, INPUT_PULLUP);
  
  // pin na vypnuti MPPT, reset vyroby v noci
  pinMode(POWEROFFMPPT_IO, INPUT_PULLUP);

  // rele spinajici napajeni
  pinMode(FVEPOWER_IO,  OUTPUT);
  pinMode(FVEPOWER_LED, OUTPUT);
  pinMode(SRVPOWER_IO,  OUTPUT);
  pinMode(SRVPOWER_LED, OUTPUT);
  FVEPOWER_OFF;
  SRVPOWER_OFF;

  // nastaveni PWM boardu
  driverPCA.begin();
  driverPCA.setPWMFreq(60);

  /*driverPCA.setPWM( 15, 0, 336 );
    while(1) { wdt_reset(); }*/

  // nastaveni poslednich poloh serv z EEPROM
  servoInit();

}



// -- loop funkce --
void loop() {

  // reset psa
  wdt_reset();

  unsigned long currentTime = millis();
  if ( currentTime < lastTime10ms ) lastTime10ms = currentTime;

  // obsluha seriove linky
  comm();


  // krok po 10-ti ms
  if ( (currentTime - lastTime10ms) > 10  ) {
    lastTime10ms = currentTime;

    // detekce signalu na zazazani prepnuti na FVE (kvuli noci)
    if (digitalRead(DISFVE_IO) == LOW) {
      if (fveSwitchEnabled == true) Serial.println("fveSwitchEnabled: prepnuti zakazano");
      fveSwitchEnabled = false;
    } else {
      if (fveSwitchEnabled == false) Serial.println("fveSwitchEnabled: prepnuti povoleno");
      fveSwitchEnabled = true;
    }

    // detekce signalu na vypnuti mppt, reset vyroby v noci
    if (digitalRead(POWEROFFMPPT_IO) == LOW) {
      if (fvePowerOff == false) {
        Serial.println("fvePowerOff: vynucene vypnuti FVE aktivovano, reset vyroby -> FVEPOWER_OFF");
        fvePowerOff = true;
      }
      FVEPOWER_OFF;
    } else {
      if (fvePowerOff == true) {
        Serial.println("fvePowerOff: vynucene vypnuti FVE zruseno -> state=INIT");
        fvePowerOff = false;
        state = INIT;
      }
    }

    TRES p = getServoChange();

    switch (state) {

      case INIT:
        Serial.println("Stav: Ininicalizace a zapnuti FVE");
        FVEPOWER_OFF;
        SRVPOWER_OFF;
        stateCounter = 400;
        state = STBY;
        break;

      case STBY:
        if (p.foundRequest) {
          srv[p.servoID].moving = true;
          state = STATE1;
          Serial.println("Stav: Pozadavek na prepnuti");
        } else {
          if ( !stateCounter-- ) {
            SRVPOWER_OFF;
            FVEPOWER_ON;
          }
        }
        break;

      case STATE1:
        Serial.println("Stav: Vypnuti FVE");
        FVEPOWER_OFF;
        stateCounter = 300;
        state = STATE2;
        break;

      case STATE2:
        if (!stateCounter--) {
          state = STATE3;
        }
        break;

      case STATE3:
        Serial.println("Stav: Zapnuti napajeni servomotoru");
        enableServo();
        stateCounter = 300;
        state = STATE4;
        break;

      case STATE4:
        if (!stateCounter--) {
          state = STATE5;
          Serial.println("Stav: Zahajeni pohybu");
        }
        break;

      case STATE5:
        if (moveServo() == true) {
          stateCounter = 300;
          state = STATE6;
          Serial.println("Stav: Ukonceni pohybu");
        }
        break;

      case STATE6:
        if (p.foundRequest) {
          srv[p.servoID].moving = true;
          state = STATE3;
          Serial.println("Stav: Dalsi pozadavek na prepnuti");
        }
        if (!stateCounter--) {
          state = STATE7;
        }
        break;

      case STATE7:
        Serial.println("Stav: Vypnuti napajeni servomotoru");
        SRVPOWER_OFF;
        stateCounter = 300;
        state = STATE8;
        break;

      case STATE8:
        if (!stateCounter--) {
          stateCounter = 100;
          state = STBY;
          Serial.println("Stav: Ukonceni procesu prepnuti a zapnuti FVE");
        }
        break;

      default:
        break;
    }
    // -- /switch --

  }
  // -- /10ms step --

}




// -- nastavi prostredi pro pohyb serva --
void enableServo() {

  unsigned char movingServoID = POCETSRV;
  // vyhledeni moving serva
  while (movingServoID--) {
    if (srv[movingServoID].moving == true) break;
  }

  // nastaveni hodnot ve strukture serva
  if (srv[movingServoID].poloha == GRID) {
    srv[movingServoID].uhel = srv[movingServoID].uhel_grid;
  }
  if (srv[movingServoID].poloha == FVE)  {
    srv[movingServoID].uhel = srv[movingServoID].uhel_fve;
  }
  if (srv[movingServoID].poloha_poz == GRID) {
    srv[movingServoID].uhel_poz = srv[movingServoID].uhel_grid;
  }
  if (srv[movingServoID].poloha_poz == FVE)  {
    srv[movingServoID].uhel_poz = srv[movingServoID].uhel_fve;
  }

  // nastaveni PWM odpovidajici aktualni poloze serva
  driverPCA.setPWM(srv[movingServoID].assoc, 0, srv[movingServoID].uhel);

  // zapnuti napajeni
  SRVPOWER_ON;

}




// -- pohyb serva mezi polohami s prestavkou v NC poloze --
unsigned char moveServo() {

  unsigned char movingServoID = POCETSRV;
  // vyhledeni moving serva
  while (movingServoID--) {
    if (srv[movingServoID].moving == true) break;
  }

  // pauza v pohybu pokud servo dojede na polohu NC
  if ((srv[movingServoID].uhel == srv[movingServoID].uhel_nc) && (srv[movingServoID].counter_nc > 0)) {
    srv[movingServoID].counter_nc--;
    return (false);
  }

  // pohyby vlevo a vpravo
  if (srv[movingServoID].uhel > srv[movingServoID].uhel_poz) {
    srv[movingServoID].uhel--;
  }
  if (srv[movingServoID].uhel < srv[movingServoID].uhel_poz) {
    srv[movingServoID].uhel++;
  }

  // zapis do PWM mutplexeru
  driverPCA.setPWM(srv[movingServoID].assoc, 0, srv[movingServoID].uhel);

  // konstanta na prodlevu v NC poloze
  srv[movingServoID].counter_nc = 200;

  // dosazeni cilove polohy
  if (srv[movingServoID].uhel == srv[movingServoID].uhel_poz) {
    srv[movingServoID].poloha = srv[movingServoID].poloha_poz;
    EEPROM.update( EEADDR + movingServoID, srv[movingServoID].poloha);
    srv[movingServoID].moving = false;
    return (true);
  }

  return (false);

}




// -- vychozi nastaveni struktury serva --
void servoInit() {

  unsigned char servoID = POCETSRV, eepoloha;
  SRVPOWER_ON;
  FVEPOWER_OFF;
  delay(3000);
  while (servoID--) {
    eepoloha = EEPROM.read(EEADDR + servoID);
    if ( (eepoloha == GRID) || (eepoloha == FVE) ) {
      srv[servoID].poloha = eepoloha;
    } else {
      srv[servoID].poloha = GRID;
      EEPROM.update(EEADDR + servoID, GRID);
    }
    if (srv[servoID].poloha == GRID) {
      srv[servoID].uhel = srv[servoID].uhel_grid;
    }
    if (srv[servoID].poloha == FVE) {
      srv[servoID].uhel = srv[servoID].uhel_fve;
    }
    srv[servoID].uhel_poz = srv[servoID].uhel;
    driverPCA.setPWM( srv[servoID].assoc, 0, srv[servoID].uhel );
    Serial.print("Nastaveni vychozi polohy serva smycky "); Serial.println(srv[servoID].assoc);
    delay(1000);
    driverPCA.setPWM( srv[servoID].assoc, 0, 0 );
    wdt_reset();
  }
  delay(2000);
  wdt_reset();
  SRVPOWER_OFF;
  delay(3000);

}



// -- zjisteni, zda-li je vyzadovana zmena polohy nejakeho prepinace --
// -- vrati se odpoved, zda-li a soucasne id serva --
TRES getServoChange() {

  TRES retVal = {false, 0};
  unsigned char servoID = POCETSRV;
  while (servoID--) {
    if ((srv[servoID].poloha_poz != NEMENIT) && (srv[servoID].poloha_poz != srv[servoID].poloha)) {
      retVal.foundRequest = true;
      retVal.servoID = servoID;
      break;
    }
  }
  return (retVal);

}
