// ----------------------------------------------------------------------------------
// SubSketch na realizaci komunikace pres RS485 sbernici
// ----------------------------------------------------------------------------------

#define MYADDR 'S'
#define RXPIN 3
#define TXPIN 4
#define TXLED 6
#define RXBUFFERSIZE 80
#define TXBUFFERSIZE 80

char rxBuffer[RXBUFFERSIZE];
char txBuffer[TXBUFFERSIZE];
unsigned char rxPtr = 0;
unsigned char txPtr = 0;
unsigned char newMsgFlag = false;



// ---------------------------------------------------------------------------------------------------------------------
// nastaveni komunikace
// ---------------------------------------------------------------------------------------------------------------------
void commSetup() {
  Serial.begin(9600, SERIAL_8O1);
  //Serial.begin(115200);
  pinMode(0, INPUT);  digitalWrite(0, HIGH); // pullup na RXD
  pinMode(TXLED, OUTPUT);
  pinMode(RXPIN, OUTPUT);
  pinMode(TXPIN, OUTPUT);
  digitalWrite(RXPIN, LOW);
  digitalWrite(TXPIN, LOW);
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// zapis prijatych dat do rx bufferu
// ---------------------------------------------------------------------------------------------------------------------
void serialEvent() {
  char znak;
  char cnt = 0;
  while (Serial.available()) {
    znak = Serial.read();
    cnt++;
    if ( !newMsgFlag && (znak > 32) && (rxPtr < RXBUFFERSIZE) ) {
      if (znak == '*') rxPtr = 0;
      rxBuffer[rxPtr] = znak;
      if ( (rxBuffer[0] == '*') && (rxBuffer[1] == MYADDR) && (isAdrValid(rxBuffer[2])) && (rxBuffer[rxPtr] == '#') ) {
        newMsgFlag = true;
      }
      rxPtr++;
    }
  }
  //Serial.println(rxPtr, DEC);
  //Serial.println(rxBuffer);
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// prevede znak 0-9A-F na cislo 0-15
// ---------------------------------------------------------------------------------------------------------------------
unsigned char ZnakToNibble(unsigned char znak) {
  if (znak >= 'A') znak = znak - 7;
  return ((znak - '0') & 0x0F);
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// prevede cislo 0-15 na znak 0-9A-F
// ---------------------------------------------------------------------------------------------------------------------
unsigned char NibbleToZnak(unsigned char nibble) {
  unsigned char retval = '0' + (nibble & 0x0F);
  if (retval >= 58) retval += 7;
  return (retval);
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// overi spravnost adresy
// ---------------------------------------------------------------------------------------------------------------------
unsigned char isAdrValid(unsigned char addr) {
  if ((addr < 'A') || (addr > 'Z')) return (false);
  return (true);
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// prida znak do tx bufferu
// ---------------------------------------------------------------------------------------------------------------------
void txBufferAdd(unsigned char znak) {
  txBuffer[txPtr++] = znak;
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// odesle paket
// ---------------------------------------------------------------------------------------------------------------------
void send(unsigned char adr) {
  digitalWrite(TXLED, HIGH);
  digitalWrite(RXPIN, HIGH);
  digitalWrite(TXPIN, HIGH);
  Serial.write('*');
  Serial.write(adr);
  Serial.write(MYADDR);
  Serial.write(txBuffer, txPtr);
  Serial.write('#');
  Serial.flush();
  digitalWrite(RXPIN, LOW);
  digitalWrite(TXPIN, LOW);
  digitalWrite(TXLED, LOW);
  txPtr = 0;
  Serial.println();
}
// ---------------------------------------------------------------------------------------------------------------------



// ---------------------------------------------------------------------------------------------------------------------
// zpracovani prijatych paketu
// ---------------------------------------------------------------------------------------------------------------------
void comm() {
  unsigned char i, c, novaPoloha, servoID, releID;

  if (newMsgFlag) {
    Serial.write(rxBuffer, rxPtr); Serial.println();

    // -- pozadavek na nastaveni switche  ---------------------------------------------
    // paket -> *SMnnnnnnnnnnnn#, kde n jsou rele 11-0, n=0 nemenit, n=1 GRID, n=2 FVE
    //       <- *MSnnnnnnnnnnnn#, kde n jsou rele 11-0, n=0 neresi,  n=1 GRID, n=2 FVE
    if ( rxPtr == 16 ) {
      i = 2;
      while (i++ < 14) {
        novaPoloha = ZnakToNibble(rxBuffer[i]);
        if (!(novaPoloha == GRID || novaPoloha == FVE)) continue;
        if ((novaPoloha == FVE) && (fveSwitchEnabled == false)) continue;
        releID = 14 - i;
        servoID = POCETSRV;
        while (servoID--) {
          if ((srv[servoID].assoc == releID) && (srv[servoID].moving == false)) {
            srv[servoID].poloha_poz = novaPoloha;
          }
        }
      }
      releID = 12;
      while (releID--) {
        c = '0';
        servoID = POCETSRV;
        while (servoID--) {
          if (srv[servoID].assoc == releID) {
            c = NibbleToZnak(srv[servoID].poloha);
            break;
          }
        }
        txBufferAdd(c);
      }
      send(rxBuffer[2]);
    }
    // --------------------------------------------------------------------------------

    newMsgFlag = false;
  }
}
// ---------------------------------------------------------------------------------------------------------------------
